/*
*	name: formparse
*	function: decode a form-data and store the result in the file
*	usage:
*		formparse [-r] [-s qstring] -o dir
*		formparse -r -s qstring -o dir   # form-data is given by qstring
*		or
*		formparse -r -o dir	# form-data is given by stdin
*		---
*		option -r: removes CR codes from the data
*
*	result:
*		decoded data is stored in files. the direcory structure is:
*		name/N	where N is 0, 1, ...
*		the count of duplicated name is in
*		name/clone
*
*	example1: in case of get-method
*		{ramfs >[2=1]} >/dev/null
*		formparse -r -s $QUERY_STRING -o /tmp
*
*	example2: in case of post-method
*		{ramfs >[2=1]} >/dev/null
*		formparse -r -o /tmp
*
*	auther: Kenji Arisawa <arisawa@aichi-u.ac.jp>
*/

#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ctype.h>

enum{
	EOD	= -1,
};

void error(char *fmt,...);
int x2i(int i);
int parse(void);
void mparse(char *dir);
int getmethod_getch(void);
int postmethod_getch(void);
int writefile(char *fn, char *b, int n);
int readfile(char *fn, char *b, int n);
int (*getch)(void);
shapeup= 0;
char *qs = 0; /* to be used for getmethod */

void
usage(void)
{
	fprint(2,"usage: formparse [-s querystring] [ -r] -o dir\n");
	exits("usage");
}

void
error(char *fmt,...)
{
	char buf[1024];
	va_list arg;
	va_start(arg, fmt);
	doprint(buf, buf+sizeof(buf), fmt, arg);
	va_end(arg);
	print("HTTP/1.0\nContent-type: text/html\n\n");
	print("<html><head><title>Form processing error</title></head>");
	print("<body><h1>Error</h1><b>%s</b></body></html>\n", buf);
	exits("error");
}

void
main(int argc, char *argv[])
{
	char *o = 0;

	ARGBEGIN{
	case 'o':
		o=ARGF();
		if(! o) usage();
		break;
	case 's':
		qs= ARGF();
		break;
	case 'r':
		shapeup= 1;
		break;
	default:
		usage();
	}ARGEND
	if(!o)
		usage();
	getch = postmethod_getch;
	if(qs)
		getch = getmethod_getch;
	mparse(o);
}

/*	conver hex char to int	*/
int
x2i(int i)
{
	i=toupper(i);
	i = i - '0';
	if(i > 9) i = i - 'A' + '9' + 1;
	return i;
}

int
getmethod_getch(void)
{
	if(*qs == 0) return -1;
	return 0xff & *qs++;
}

int
postmethod_getch(void)
{
	char buf[1];
	/*	don't use bio. getch may be hung
	*	EOD of postmethod is '\r' 		*/
	if(read(0,buf,1) != 1) return -1;
	if(buf[0] == '\r') return -1;
	return (unsigned) buf[0];
}

int
readfile(char *fn, char *b, int n)
{
	int fd;
	fd = open(fn, OREAD);
	if(fd < 0)
		return -1;
	n = read(fd, b, n);
	b[n] = 0;
	close(fd);
	return n;
}

int
writefile(char *fn, char *b, int n)
{
	int fd;
	fd = open(fn, OWRITE);
	if(fd < 0)
		return -1;
	n = write(fd, b, n);
	close(fd);
	return n;
}


/*
*	parse single name and the value
*	return:
*	0:	parsing single name/value pair has finished
*	-1:	End of data or fatal error
*/
int
parse(void)
{
	char *s, *s0;
	char name[NAMELEN], buf[256], fn[NAMELEN*2];
	int c, c1;
	Biobuf *bout;
	Dir dir;
	int fd, n;

	/*
	*	regular format is,
	*	name=val&name=val&...name=val
	*	but the data can be forged, so we must be careful.
	*	example of forged patterns:
	*	&&&&
	*	====
	*	=&=&=
	*/

	/*
	*	parse a name
	*	we restrict chars in the name to be alnum
	*/
	s = s0 = name;
	while((c = getch()) != EOD){
		if(s - s0 == NAMELEN - 1){ /* truncate here */
			break;
		}
		if(c == '='){
			break;
		}
		if(isalnum(c)){
			*s++ = c;
			continue;
		}
		/* something wrong. skip it */
	}
	*s = 0;
	if(c == EOD)
		return EOD;

	/* sweep out extra chars */
	while(c != '=' && (c = getch()) != EOD);

	/* check the name length */
	if(strlen(name) == 0){ /*something wrong. skip to next name */
		while((c = getch()) != EOD && c != '&');
		return 0;
	}

	/*
	*	we must count the occurrence of the name.
	*	count it in:
	*		name/clone
	*/
	if(dirstat(name, &dir) <0){ /* the name is first apeared */
		fd = create(name, OREAD, 0750 | CHDIR);
		if(fd < 0) /* somthing wrong */
			error("could not create %s", name);
		close(fd);
		snprint(fn,sizeof(fn),"%s/clone",name);
		fd = create(fn, OWRITE, 0640);
		if(fd < 0) /* somthing wrong */
			error("%s not open", fn);
		write(fd,"0",1);
		close(fd);
	}

	/*
	*	read clone. increase the value. write it back
	*	the name and the value is sored in:
	*	name/0, name/1, ...
	*/
	snprint(fn,sizeof(fn),"%s/clone",name);
	if(readfile(fn, buf, sizeof(buf))<0)
		error("cannot read %s", fn);

	snprint(fn,sizeof(fn),"%s/%s",name,buf);
	bout = Bopen(fn, OWRITE);
	if(!bout)
		error("%s not open", fn);
	n = atoi(buf) + 1;
	snprint(buf,sizeof(buf),"%d", n);
	snprint(fn,sizeof(fn),"%s/clone",name);
	if(writefile(fn, buf, strlen(buf)) < 0)
		error("cannot write to %s", fn);

	/* get the value */
	while((c = getch()) != EOD){
		switch(c){
		case '&':
			Bterm(bout);
			return 0;
		case '+':
			c = ' ';
			break;
		case '%':
			c = getch();
			if(c == EOD) goto err;
			c1 = c;
			c = getch();
			if(c == EOD) goto err;
			c = 16*x2i(c1) + x2i(c);
			if(c == '\r' && shapeup) continue;
			break;
		default:
			;
		}
		Bputc(bout, c);
	}
err:
	Bterm(bout);
	return EOD;
}

/*	parse multiple names	*/
void
mparse(char *dir)
{
	if(chdir(dir) < 0)
		error("could not chdir to %s",dir);
	while(parse() != EOD);
}
