#include <u.h>
#include <libc.h>
#include <auth.h>
#include <ctype.h>
#include <bio.h>
#include <ndb.h>

#include "httpd.h"
#include "httpsrv.h"

#define debug if
#define noop if(0)

typedef struct Endpoints	Endpoints;

struct Endpoints
{
	char	*lsys;
	char	*lserv;
	char	*rsys;
	char	*rserv;
};

/*	our default options	*/
static char *webroot="/usr/web";
char *rewrite = "/sys/lib/httpd.rewrite";
char *HTTPLOG = "http";
static char *namespace = "/lib/namespace.httpd";
static char *address = "tcp!*!http";
static int	server = 0; // 0: runs under the control of listen. 1: runs in server mode.
static int timeout = 5; // timeout for CGI
static int mntok=0; //should be mntok=0 in release code
char	*netdir = "/net";

/* internal params */
char *serversoft="Pegasus/0.8 (Plan9)";
static const int contmax = 100; // max continuation count for safty

/* used internally for control */
static char webmnt[256]; // to be ("%s/mnt",webroot)
static char webetc[256]; // to be ("%s/etc", webroot)
static char *myport; //  80 if you use default http port
static char *myname; // example: ar in ar.aichi-u.ac.jp
static char *mydom; // example: aichi-u.ac.jp
static int cpid; // pid of CGI
static int mnt; // namespace configuration is finished
static int cont; // Keep-Alive continuation count
static int notestatus; // alarm note to CGI is rceived

/*	Note on cont
*	cont is used as a flag so that we can detect re-request using Keep-Alive
*	if (cont > 0)
*		re-request due to Kepp-Alive
*		so we should not reconfigure our namespace
*		cont means continued
*/

static	void		becomenone(void);
static	char*	csquery(char*, char*, char*);
static	void		dolisten(char*);
static	int		doreq(HConnect*);
static	int		send(HConnect*);
static	char*		sysdom(void);
static	Endpoints*	getendpoints(char*);
static	void		eputenv(char *name, char *value);
static 	void		doservice(Endpoints *end);
static 	int		execute(HConnect *c, char *w);
static	int		executable(char *path);
static	int		waitfor(int pid, char *msg);
static	int		hparse2env(HConnect *c, char *head);
static	char*	trim(char *s);
static 	int		notefun(void *a, char *msg);
static 	int	execsafely(HConnect *c, char *path, char **argv);
static	int		merge(char *mnt, char *root);
static	int	ipcheck(HConnect *c, char *file, char *ip);
static 	char*	redirect(HConnect *c, char *file, char *path);
static	void		setcgienv(HConnect *c);

static void
usage(void)
{
	fprint(2, 
	"usage: httpd [-sm] [-t timeout] [-a srvaddress] [-d domain] [-n namespace] [-r webroot] [-R rewrite] [[httpd] dummy ..  incalldir]\n");
	exits("usage");
}

void
main(int argc, char **argv)
{
	char *incalldir = 0;
	hmydomain = nil;
	fmtinstall('D', hdateconv);
	fmtinstall('H', httpconv);
	fmtinstall('U', hurlconv);
	ARGBEGIN{
	case 'n':
		namespace = ARGF();
		break;
	case 'a':
		address = ARGF(); /* tcp!*!http */
		break;
	case 'd':
		hmydomain = ARGF();
		break;
	case 't':
		timeout = atoi(ARGF());
		break;
	case 'r':
		webroot = ARGF();
		break;
	case 'R':
		rewrite = ARGF();
		break;
	case 's':
		server = 1;
		break;
	case 'm':
		mntok = 1;
		break;
	default:
		usage();
		break;
	}ARGEND

	if(argc > 0)
		incalldir = argv[argc-1];
	if((server && incalldir) || (!server && !incalldir))
		usage();

	/* cleanname strip out tailing '/'	*/
	cleanname(webroot);

	if(webroot[0] != '/')
		sysfatal("webroot must begin with `/' ");
	if(strcmp(webroot, "/") == 0)
		sysfatal("webroot `/' disallowed");
	if(namespace[0] != '/')
		sysfatal("namespace must begin with `/'");
	if(rewrite[0] != '/')
		sysfatal("rewrite must begin with `/'");
	if(hmydomain == nil)
		hmydomain = sysdom();
	/* extract mydom */
	myname = sysname();
	if(strlen(hmydomain) < strlen(myname)){
		/* something wrong */
		syslog(0, HTTPLOG, "invalid sysname and sysdom: %s %s", myname, hmydomain);
		exits("sysname > sysdom");
	}
	mydom = hmydomain + strlen(myname) + 1;

	snprint(webmnt, sizeof webmnt, "%s/mnt", webroot);
	snprint(webetc, sizeof webetc, "%s/etc", webroot);

	if(!server){
		Endpoints *ends;
		time(nil);
		contentinit();
		urlinit();
		statsinit();
		ends = getendpoints(incalldir);
		doservice(ends);
		exits(nil);
	}

	switch(rfork(RFNOTEG|RFPROC|RFFDG|RFNAMEG)) {
	case -1:
		sysfatal("fork");
	case 0:
		break;
	default:
		exits(nil);
	}

	/*
	 * open all files we might need before castrating namespace
	 */
	time(nil);
	syslog(0, HTTPLOG, nil);

	contentinit();
	urlinit();
	statsinit();

	becomenone();
	dolisten(address);
	exits(nil);
}

static void
becomenone(void)
{
	int fd;

	fd = open("#c/user", OWRITE);
	if(fd < 0 || write(fd, "none", strlen("none")) < 0)
		sysfatal("can't become none");
	close(fd);
	if(newns("none", nil) < 0)
		sysfatal("can't build normal namespace");
}

static HConnect*
mkconnect(void)
{
	HConnect *c;

	c = ezalloc(sizeof(HConnect));
	c->hpos = c->header;
	c->hstop = c->header;
	c->replog = 0;
	return c;
}

static HSPriv*
mkhspriv(void)
{
	HSPriv *p;

	p = ezalloc(sizeof(HSPriv));
	return p;
}

static void
dolisten(char *address)
{
	Endpoints *ends;
	char ndir[NETPATHLEN], dir[NETPATHLEN], *p;
	int ctl, nctl, data, spotchk;
	char		netdirb[256];

	spotchk = 0;
	syslog(0, HTTPLOG, "httpd starting");
	ctl = announce(address, dir);
	if(ctl < 0){
		syslog(0, HTTPLOG, "error: can't announce on %s: %r", address);
		return;
	}

	/* 	the logic bellow seems curious to me -Kenar- */
	strcpy(netdirb, dir);
	p = nil;
	if(netdir[0] == '/'){ // netdirb !? -Kenar-
		p = strchr(netdirb+1, '/');
		if(p != nil)
			*p = '\0';
	}
	if(p == nil)
		strcpy(netdirb, "/net");
	netdir = netdirb;

	for(;;){

		/*
		 *  wait for a call (or an error)
		 */
		nctl = listen(dir, ndir);
		if(nctl < 0){
			syslog(0, HTTPLOG, "error: can't listen on %s: %r", address);
			syslog(0, HTTPLOG, "error+ ctls = %d", ctl);
			return;
		}

		/*
		 *  start a process for the service
		 */
		switch(rfork(RFFDG|RFPROC|RFNOWAIT|RFNAMEG)){
		case -1:
			close(nctl);
			continue;
		case 0:
			/*
			 *  see if we know the service requested
			 */
			data = accept(ctl, ndir);
			if(data < 0){
				syslog(0, HTTPLOG, "error: can't open %s/data: %r", ndir);
				exits(nil);
			}
			dup(data, 0);
			dup(data, 1);
			dup(data, 2);
			close(data);
			close(ctl);
			close(nctl);

			ends = getendpoints(ndir);
			doservice(ends);

			exits(nil);
		default:
			close(nctl);
			break;
		}

		if(++spotchk > 50){
			spotchk = 0;
			contentinit();
			urlinit();
			statsinit();
		}
	}
}

static void
doservice(Endpoints *ends)
{
	HSPriv *hp;
	HConnect *c;
	Ndb *db;
	Ndbs s;
	Ndbtuple *nt;
	char buf[Ndbvlen];
	int t, ok;

	c = mkconnect();
	hp = mkhspriv();
	hp->remotesys = ends->rsys;
	hp->remoteserv = ends->rserv;
	myport=ends->lserv;
	c->private = hp;

	db = ndbopen(nil);
	if(db == 0){
		syslog(0, HTTPLOG, "error: no db files");
		exits("no db");
	}
	/*	we pic up only one	*/
	nt = ndbgetval(db, &s, "ip", ends->lsys, "sys", buf); 
	if(nt)
		ndbfree(nt);
	ndbclose(db);
	if(!nt){
		syslog(0, HTTPLOG, "error: no sys name in ndb file");
		exits("no sysname");
	}
	//syslog(0, HTTPLOG, "debug: ndb query %s %s", ends->lsys, buf);
	myname = strdup(buf);

	hinit(&c->hin, 0, Hread);
	hinit(&c->hout, 1, Hwrite);

	/*
	 * serve requests until a magic request.
	 * later requests have to come quickly.
	 * only works for http/1.1 or later.
	 */
	for(t = 15*60*1000; ; t = 15*1000){
		if(hparsereq(c, t) <= 0)
			exits(nil);
		ok = doreq(c);

		if(c->head.closeit || ok < 0)
			exits(nil);

		hreqcleanup(c);
	}
}

static int
doreq(HConnect *c)
{
	char *uri, *origuri, *p, *host;
	char virtualhost[100];
	char *urioptions;

	uri = c->req.uri;
	/*
	 * normal case is just file transfer
	 */
	if(cont > 0){
		logit(c, "continue: %d uri=%s",cont, uri);
		if(cont > contmax){ /* may be required for safty */
			logit(c, "continuation exceed max: %d uri=%s",cont, uri);
			return hfail(c, HInternal);
			exits(nil);
		}
		cont++;
	}

	if(hparseheaders(c, 15*60*1000) < 0)
		exits("failed");
	if(!http11(c) && !c->head.persist)
		c->head.closeit = 1;

	{	/*	logit	*/
		char *p1,*p2;
		int n1 = 0;
		int n2 = 0;
		int n;
		p1 = p2 = nil;
		if(c->req.params)
			p1 = c->req.params;
		if(c->req.search)
			p2 = c->req.search;
		if(p1)
			n1 = strlen(p1);
		if(p2)
			n2 = strlen(p2);
		n = n1 + n2 + 3 + UTFmax;
		urioptions = halloc(c, n);
		if(p1 && p2)
			snprint(urioptions, n, ";%s?%s", p1, p2);
		else if(p1)
			snprint(urioptions, n, ";%s", p1);
		else if(p2)
			snprint(urioptions, n, "?%s", p2);
		else
			strcpy(urioptions,"");
		logit(c, "%s %s%s", c->req.meth, uri, urioptions);
	}

	debug(0){
		int ch;
		ch = *(c->hstop);
		*(c->hstop) = 0;
		logit(c,"%s", (char*)c->header);
		*(c->hstop) = ch;
	}

	/*	Let ar be a host name and aichi-u.ac.jp is the dom
	*	then head.host is:
	*	ar
	*	ar.aichi-u.ac.jp
	*	ar.aichi-u.ac.jp:80	# some browsers send GET header with :80
	*	-Kenar-	*/
	host = c->head.host;
	/*	we can remove `:80' 	here */
	p = strchr(host,':');
	if(p)
		*p = 0;

	/*	we can remove mydom part from head.host
	*	so that we can simplify management of
	*	/sys/lib/httpd.rewrite.
	*	now we don't need full name for virtual host in this file.
	*	-Kenar-	*/
	if(strlen(host)>strlen(mydom)){
		char buf[256];
		snprint(buf, sizeof buf, ".%s", mydom);
		p = host + strlen(host) - strlen(buf);
		if(strcmp(p, buf) == 0)
			*p = 0;
	}

	snprint(virtualhost, sizeof virtualhost, "http://%s/", c->head.host);
	origuri = uri;
	uri = redirect(c, rewrite, virtualhost);
	/*
	*	we can reject some faked accesses here;
	*	if uri is nil and host is not myname, we can reject.
	*	example:
	*	ar	# sysname
	*	plan9	# another name of this host ar
	*	aichi-u.ac.jp	# mydom
	*	now ar.aichi-u.ac.jp and plan9.aichi-u.ac.jp are reduced to
	*	ar and plan9 respectively.
	*	On the other hand
	*	www.faked.com
	*	is keep untached.
	*/
	if(uri == nil){
		if(strcmp(host, myname) != 0){
			logit(c, "denied: header host %s", host);
			return hfail(c, HNotFound);
		}
		uri = redirect(c, rewrite, origuri);
	}

	debug(0)
		logit(c,"redirect: new uri is %s", uri);

	mnt = 0;
	if(uri && *uri == '*'){
		//	This case is '*' redirect in /sys/lib/httpd.rewrite such as:
		//	http://pegasus  */usr/someone/www
		if(!cont){
			int status;
			status = bind(uri + 1, webmnt, MREPL);
			if(status < 0){
				logit(c,"error: bind %s %s # %r", uri+1,webmnt);
				return hfail(c, HNotFound);
			}
			mnt = 1;
		}
	}
	else if(uri != nil){ /* Let the client know the new address */
		char *newuri;
		int n;
		n = strlen(uri) + strlen(urioptions) + 1 + UTFmax;
		newuri = halloc(c, n);
		snprint(newuri, n, "%s%s", uri, urioptions);
		return hmoved(c, newuri);
	}


	return send(c);
}

/*	# return values
*	+1: OK
*	-1: NG
*	-Kenar-	*/
static int
send(HConnect *c)
{
	Dir dir;
	char *w, *p, *suffix;
	char *uri, *newuri;
	char *path = 0;
	char *user = 0;
	int fd, fd1, n, ok;
	char buf[256];
	int status;
	HSPriv *hp = c->private;
	USED(user);

	/* 	Here we support http://ar/~alice/abc
		then we bind /usr/alice/web to /usr/web/mnt
		and change request document to /abc
		This redirection is automatically done without /sys/lib/httpd.rewrite
		-Kenar-
	 */
	uri = c->req.uri;
	if(mnt == 0 && uri[0]=='/' && uri[1]=='~'){
		char *t;
		t = strchr(uri+2, '/');

		/* 	if uri is "/~alice" for example,
			let the client resend the correct uri "/~alice/"
			-Kenar-	*/
		if(!t){
			snprint(buf, sizeof buf, "%s/", uri);
			return hmoved(c, buf);
		}

		if(!cont){
			user = ezalloc(t -uri+1); // user is "alice"
			strncpy(user, uri+2,  t - uri - 2);
			c->req.user = user;
			n = strlen("/usr//web") + strlen(uri) + 1 + UTFmax;
			path = halloc(c, n);
			snprint(path, n, "/usr/%s/web", user);
			status = bind(path, webmnt, MREPL);
			if(status == -1){
				logit(c, "error: bind %s %s # %r", path, webmnt);
				return notfound(c, uri+1);
			}
			mnt = 1;
		}
		path=t;

		/* 	then c->req.uri is "/~alice/abc", and path is "/abc" 
			for the request http://ar/~alice/abc
			We need not be wary of "/~alice/../",
			it reduces to "/" as it should be.
			-Kenar-	*/
	}
	if(!path) path = uri;
	c->req.path = path;

	/* namespace reconfiguration; hide everything out of webroot */
	if(!cont && addns("none", namespace) < 0){
		logit(c, "error: addns %s # %r", namespace);
		sysfatal("can't build httpd namespace");
	}

	if(!cont){
		if(mnt){
			unmount(nil, webetc);
			status = merge(webmnt,webroot);
			if(status < 0){
				logit(c, "error: merge %s %s # %r", webmnt, webroot);
				return hfail(c, HInternal);
			}
			status = unmount(nil, webmnt);
			if(status < 0){
				logit(c, "error: unmount %s # %r", webmnt);
				return hfail(c, HInternal);
			}
		}

		status = bind(webroot,"/", MREPL);
		if(status == -1){
			logit(c, "error: bind %s / # %r", webroot);
			logit(c, "error+ header=%s",(char*)c->header);
			return hfail(c, HInternal);
		}
		cont = 1;
	}
	chdir("/");

	if(strcmp(c->req.meth, "GET") != 0
		&& strcmp(c->req.meth, "HEAD") != 0
		&& strcmp(c->req.meth,"POST") != 0)
		return hunallowed(c, "GET, HEAD, POST");
	if(c->head.expectother || c->head.expectcont)
		return hfail(c, HExpectFail);

	newuri = redirect(c, "/etc/rewrite", path);
	if(newuri)
		return hmoved(c, newuri);

	if(ipcheck(c, "/etc/allow", hp->remotesys) < 0)
		return hfail(c, HUnauth);

	ok = authcheck(c, "/etc/passwd");
	if(ok <= 0)
		return ok;


	/*
	 * check for directory/file mismatch with trailing /,
	 * and send any redirections.
	 */
	n = strlen("/doc") + strlen(path) + strlen("/index.html") + 1 + UTFmax;
	/* make a room to insert index.html */

	w = halloc(c, n);
	strcpy(w, "/doc");
	strcat(w,path);

	fd = open(w, OREAD);
	if(fd < 0){
		logit(c, "error: open %s # %r", w);
		return notfound(c, uri);
	}
	if(dirfstat(fd, &dir) < 0){
		close(fd);
		logit(c, "error: %s internal error", uri);
		return hfail(c, HInternal);
	}
	p = strchr(w, '\0');
	if(dir.mode & CHDIR){
		/*	if uri is "path" and the path is a direcotory
			then let the client resent "path/"
			-Kenar-	*/
		if(uri[strlen(uri)-1] != '/'){
			snprint(buf, sizeof buf, "%s/", uri);
			return hmoved(c, buf);
		}
		
		/*	here the form of uri is "path/"
			therefore p must be > 0 , and w ends with '/'
			-Kenar-		*/
		strcat(w, "index.html");

		fd1 = open(w, OREAD);
		if(fd1 < 0){
			close(fd);
			logit(c, "%s not found", uri);
			return notfound(c, uri);
		}
		close(fd);
		fd = fd1;
		if(dirfstat(fd, &dir) < 0){
			close(fd);
			logit(c, "error: %s internal error", uri);
			return hfail(c, HInternal);
		}
	}else if(p > w && p[-1] == '/'){
		/*	This condition means:
			w is a file and ends with '/' 
			for example,
				http://ar.aichi-u.ac.jp/index.html/
			Then let the client to resend
				http://ar.aichi-u.ac.jp/index.html
			-Kenar-	*/
		close(fd);
		*strrchr(c->req.uri, '/') = '\0';
		logit(c, "%s hmoved", uri);
		return hmoved(c, uri);
	}

	/*	Let our sendfd() know the contents type to assign full pathname
	*	to req.uri
	*	-Kenar-	*/
	c->req.uri = w + strlen("/doc");

	/*	Now every redirection is finished and we have ready.
	*	requested path is already converted to absolute path.
	*	values to be used are:
	*	w: path to the file to send. path is in httpd name space
	*	fd: open file descriptor of w
	*
	* 	CGI is supported. The name must be one of the form:
	*	*.html *.cgi *.http
	*	and they must be exec permitted
	*	-Kenar-
	*/

	/* we disallow files starting with `.' */
	p = strrchr(w,'/');
	if(p[1] == '.'){
		close(fd);
		logit(c, "%s not found", uri);
		return notfound(c, uri);
	}

	suffix = strrchr(w,'.');
	if(suffix
	&& (strcmp(suffix,".html")==0 || strcmp(suffix,".cgi") == 0 || strcmp(suffix,".http") == 0)
	&& executable(w)){
		/*	Don't use access(w, AEXEC)==0 instead of executable(w)
		*	Why? If the fs is running in allow mode, all files are
		*		access(w, AEXEC)==0
		*	-Kenar-	*/

		if(strcmp(c->req.meth,"POST") == 0){
			int fd[2];
			char *hb;
			hb = hunload(&c->hin);
			if(hb == nil)
				hfail(c, HInternal);
			/* the result is buggy, so we must skip two chars. */
			hb += 2;

			/* pass the hb to exec() */
			pipe(fd);
			switch(fork()){
			case -1:
				logit(c, "error: fork # %r");
				hfail(c, HInternal);
				return -1;
			case 0:
				/*	Let the child exit, because I would like to let CGI
				*	succeed to this pid
				*	-Kenar-	*/
				close(0);
				close(fd[0]);
				dup(fd[1], 1);
				close(fd[1]);
				/*	POST data have been received in hb,
				*	so we pass the data to the child process */
				print("%s\n", hb);
				exits(nil);
			default:
				close(fd[1]);
				dup(fd[0], 0);
				close(fd[0]);
				break;
			}
		}

		if(strcmp(suffix,".http") != 0){
			Hio *hout = &c->hout;
			hprint(hout, "%s 200 OK\r\n", hversion);
			hprint(hout, "Server: %s\r\n", serversoft);
			hprint(hout, "Date: %D\r\n", time(nil));
			if(strcmp(suffix,".html") == 0){
				hprint(hout, "Content-Type: text/html\r\n");
				if(c->head.closeit)
					hprint(hout, "Connection: close\r\n");
				else if(!http11(c))
					hprint(hout, "Connection: Keep-Alive\r\n");
				hprint(hout, "\r\n");
			}
			hflush(hout);
		}

		setcgienv(c);
		return execute(c,w);
	}
	status = sendfd(c, fd, &dir, nil, nil);
	if(status == -1)
		logit(c, "error: sendfd %s # %r", w);
	else
		logit(c, "sent");
	return status; // +1: success;  -1: fail
}

static char*
sysdom(void)
{
	char *dn;

	dn = csquery("sys" , sysname(), "dom");
	if(dn == nil)
		dn = "who cares";
	return dn;
}

/*
 *  query the connection server
 */
static char*
csquery(char *attr, char *val, char *rattr)
{
	char token[64+4];
	char buf[256], *p, *sp;
	int fd, n;

	if(val == nil || val[0] == 0)
		return nil;
	snprint(buf, sizeof(buf), "%s/cs", netdir);
	fd = open(buf, ORDWR);
	if(fd < 0)
		return nil;
	fprint(fd, "!%s=%s", attr, val);
	seek(fd, 0, 0);
	snprint(token, sizeof(token), "%s=", rattr);
	for(;;){
		n = read(fd, buf, sizeof(buf)-1);
		if(n <= 0)
			break;
		buf[n] = 0;
		p = strstr(buf, token);
		if(p != nil && (p == buf || *(p-1) == 0)){
			close(fd);
			sp = strchr(p, ' ');
			if(sp)
				*sp = 0;
			p = strchr(p, '=');
			if(p == nil)
				return nil;
			return estrdup(p+1);
		}
	}
	close(fd);
	return nil;
}


/*	dir is "/net/tcp/5" for example
*	file is "remote", "local"
*	 -Kenar-	*/
static void
getendpoint(char *dir, char *file, char **sysp, char **servp)
{
	int fd, n;
	char buf[128];
	char *sys, *serv;

	sys = serv = nil;

	snprint(buf, sizeof buf, "%s/%s", dir, file); 
	fd = open(buf, OREAD);
	if(fd >= 0){
		n = read(fd, buf, sizeof(buf)-1);
		/* then we will read "202.250.160.40!80" for success */
		if(n>0){
			buf[n-1] = 0;
			serv = strchr(buf, '!');
			if(serv){
				*serv++ = 0;
				serv = estrdup(serv);  // "80" for example
			}
			sys = estrdup(buf); // "202.250.160.40" for example
		}
		close(fd);
	}
	if(serv == nil)
		serv = estrdup("unknown");
	if(sys == nil)
		sys = estrdup("unknown");
	*servp = serv;
	*sysp = sys;
}

static Endpoints*
getendpoints(char *dir)
{
	Endpoints *ep;

	ep = ezalloc(sizeof(*ep));
	getendpoint(dir, "local", &ep->lsys, &ep->lserv);
	getendpoint(dir, "remote", &ep->rsys, &ep->rserv);
	return ep;
}

/*------- the bellow is my addition ----Kenar----*/

static void
eputenv(char *name, char *value)
{
	if(value) putenv(name,value);
}

static void
setcgienv(HConnect *c)
{
	HSPriv *hp = c->private;
	char buf[256];
	int ch;

	rfork(RFCENVG);

	/*	Putting header info into environment
	*	must be the first. (for security)
	*	-Kenar-	*/
	ch = *(c->hstop);
	*(c->hstop) = 0;
	eputenv("HTTP_HEADER", (char*)c->header);
	hparse2env(c, (char*)c->header);
	*(c->hstop) = ch;

	eputenv("DOCUMENT_ROOT", "/doc");
	eputenv("GATEWAY_INTERFACE", "CGI/1.1");

	snprint(buf, sizeof buf, "%s!%s", hp->remotesys, hp->remoteserv);
	eputenv("REMOTE_ADDR", buf);

	eputenv("REQUEST_PATH", c->req.path);
	eputenv("REQUEST_METHOD",c->req.meth);
	
	snprint(buf, sizeof buf, "%d.%d", c->req.vermaj,c->req.vermin);
	eputenv("REQUEST_VERSION", buf);
	eputenv("REQUEST_URI",c->req.uri);
	eputenv("AUTH_USER", c->head.authuser);
	eputenv("AUTH_PASS", c->head.authpass);
	eputenv("REQUEST_USER", c->req.user);

	eputenv("SERVER_DOMAIN", hmydomain);
	eputenv("SERVER_NAME", sysname());
	eputenv("SERVER_PORT", myport);
	eputenv("SERVER_PROTOCOL",hversion);
	eputenv("SERVER_SOFTWARE", serversoft);
	eputenv("QUERY_STRING", c->req.search);
	eputenv("home", "/doc");
	snprint(buf, sizeof buf, "%d", timeout);
	eputenv("timeout", buf);
	eputenv("namespace", namespace);
}

static int
execute(HConnect *c, char *w)
{
	char *p;
	char  **argv;
	char *params;
	char *basename;
	int k, n;
	char *q;
	
	p = strrchr(w,'/');
	if(p)
		*p = 0;
	if(strlen(w)>0) chdir(w);
	if(p)
		*p++ = '/';
	else
		p = w;
	basename = p;
	params = c->req.params;
	/*	params points "alice;bob"
	*	In creating argv for exec call,
	*	we should allocate memory for argv.
	*	argv[0] must be reserved for basename of exec file
	*	argv[n-1] must be nil where n is num of elements
	*	so, I use ";alice;bob;" to apply getfield.
	*	-Kenar-		*/
	if(params == nil){
		argv = halloc(c, 2*sizeof(char*));
		n = 2;
	}
	else{
		n = strlen(";;") + strlen(params) + 1 + UTFmax;
		q = halloc(c, n);
		snprint(q, n, ";%s;", params);
		params = q;
		n = 0;
		for(p = params; *p; p++)
			if(*p == ';') n++;
		argv = halloc(c, (n+1) * sizeof(char*));
		n = getfields(params, argv, n+1, 0, ";");
		/* don't forget to hurlunesc */
		for(k = 1; k < n -1; k++)
			argv[k] = hurlunesc(c, argv[k]);
		/* it is not good idea to trim extra spaces */
	}
	argv[0] = basename;
	argv[n-1] = nil;

	/*	if it takes so much time to complete, kill the process
	*	-Kenar-	*/
	return execsafely(c, w, argv);
}


#ifdef NODEF
/*	This code intends to get output of child and pass it to the client,
*	so that server can know the child worked properly.
*	This code may be helpful when we support
*	Content-Length
*	in HTTP header		*/
static int
execsafely(HConnect *c, char *path, char **argv)
{
	int fd, fd0, fd2[2], n, status;
	char buf[16*1024];

	/*	ramfs service to the child	*/
	switch(rfork(RFFDG|RFPROC)){
	case -1:
		logit(c, "error: fork # %r");
		hfail(c, HInternal);
		return -1;
	case 0:
		close(0);
		close(1);
		close(2);// prevent "ramfs: unmount read" message
		execl("/bin/ramfs","ramfs",0);
		/* not reached */
		logit(c, "error: unable to execl /bin/ramfs %r");
		exits(nil);
	default:
		break;
	}

	/* 	read the output of child CGI, wait until finished.
	*	If it takes too much time, kill the child and inform to the client  */
	pipe(fd2);
	switch(cpid = rfork(RFFDG|RFREND|RFPROC|RFNOTEG)){ /* not == */
	case -1:
		logit(c, "error: fork # %r");
		hfail(c, HInternal);
		return -1;
	case 0: // child: read stdin, redirect stdout to parent
		close(fd2[0]);
		dup(fd2[1], 1);
		close(fd2[1]);
		dup(2,1);
		break;
	default:// parent: read the output of CGI and write it to stdout
		fd0 = dup(0, -1);
		close(0);
		close(fd2[1]);
		fd = fd2[0];
		atnotify(notefun,1);
		alarm(timeout*1000);
		notestatus = 1;
		n = read(fd, buf, sizeof buf);
		while(n > 0){
			//hwrite(&c->hout, buf, n);
			write(1, buf, n);
			n = read(fd, buf, sizeof buf);
		}
		//hflush(&c->hout);
		close(fd);
		dup(fd0, 0);
		alarm(0);
		atnotify(notefun,0);
		if(notestatus != 1){
			logit(c,"alarm done for pid=%d", cpid);
			return -1;
		}
		unmount(nil,"/tmp");
		return 1;
	}

	if(!!mntok){
		status = rfork(RFNOMNT);
		if(status < 0){
			logit(c,"error: rfork nomnt # %r");
			return hfail(c, HInternal);
	}
	exec(path, argv);
	/* not reached */
	logit(c,"error: exec %s # %r", path);
	exits(nil);// don't return
}
#endif

/*	This code intends to get exit status from the child	*/
static int
execsafely(HConnect *c, char *path, char **argv)
{
	int fd, pid;
	int status;
	char msg[ERRLEN];

	/*	ramfs service to the child	*/
	switch(pid = rfork(RFFDG|RFPROC)){ /* assign = */
	case -1:
		logit(c, "error: fork # %r");
		hfail(c, HInternal);
		return -1;
	case 0:
		close(0);
		close(1);
		close(2);// prevent "ramfs: unmount read" message
		execl("/bin/ramfs","ramfs",0);
		/* not reached */
		logit(c, "error: unable to execl /bin/ramfs %r");
		exits(nil);
	default:
		break;
	}

	/* ensure "/bin/ramfs" finished */
	status = waitfor(pid, msg);
	if(status < 0){
		logit(c,"error: /bin/ramfs: no children to wait %d", pid);
		return -1;
	}

	/* 	wait child to finish.
	*	If it takes too much time, kill the child and get message  */
	switch(cpid = rfork(RFFDG|RFPROC|RFNOTEG)){ /* assign = */
	case -1:
		logit(c, "error: fork # %r");
		hfail(c, HInternal);
		return -1;
	case 0: // child: read stdin, and write to stdout
		unmount(nil,"/net");
		if(!mntok){
			status = rfork(RFNOMNT);
			if(status < 0){
				logit(c,"error: rfork nomnt # %r");
				return hfail(c, HInternal);
			}
		}
		exec(path, argv);
		/* not reached */
		logit(c,"error: exec %s # %r", path);
		exits(nil);// don't return
		break;
	default:// parent: wait for child to finish.
		break;
	}
	fd = dup(0, -1);
	close(0);
	atnotify(notefun,1);
	alarm(timeout*1000);
	notestatus = 1;
	status = waitfor(cpid, msg);
	alarm(0);
	atnotify(notefun,0);
	dup(fd, 0);
	close(fd);
	unmount(nil,"/tmp");
	if(notestatus != 1){
		logit(c,"alarm done to pid=%d", cpid);
		return -1;
	}
	if(status < 0){
		logit(c,"error: no children to wait %d", cpid);
		return -1;
	}
	if(msg[0] != 0){
		logit(c,"child %d exits with msg %s", cpid, msg);
		return -1;
	}
	logit(c, "done");
	return 1;
}


/*
*	exeecutable returns 1
*	if the file is executable for `others'
*	else returns 0
*/
static int
executable(char *path)
{
	Dir dir;
	dirstat(path, &dir);
	if(dir.mode & CHEXEC)
		return 1;
	return 0;
}

static int
waitfor(int pid, char *msg)
{	int cpid;
	Waitmsg w;
	while((cpid = wait(&w)) >= 0){
		if(pid == cpid){
			strncpy(msg, w.msg, ERRLEN);
			return 0;
		}
	}
	return -1;
}


/*	mime parse is not easy.
*	and probably meaningless to put the info int env.
*	so I chose a simple way
*	I parsed only
*		key:val
*	pairs, and
*		putenv(key, val)
*	-Kenar-	*/
static int
hparse2env(HConnect *c, char *head)
{
	char *p,*s;
	char **buf;
	char *kv[2];
	int k,n,m;

	p = hstrdup(c, head);
	/* skip a line */
	p = strchr(p, '\n');
	if(!p)
		return -1;
	p++;
	n = 0;
	for(s = p; *s; s++)
		if(*s == '\n') n++;
	buf = halloc(c, (n + 1) * sizeof(char*));
	n = getfields(p, buf, n + 1, 0, "\n");
	
	/* strip the trailing CR */
	for(k = 0; k < n; k++){
		p = strchr(buf[k], '\r');
		if(p)
			*p = 0;
	}

	/*	get
	*		key: val
	*	and trim, then putenv	*/
	for(k = 0; k < n; k++){
		m = getfields(buf[k], kv, 2, 0, ":");
		if(m != 2)
			continue;
		for(p = kv[0]; *p; p++)
			*p = toupper(*p);
		for(p = kv[0]; *p; p++)
			if(*p == '-') *p = '_';
		eputenv(trim(kv[0]), trim(kv[1]));
	}
	return 0;
}


static char*
trim(char *s)
{
	char *t;
	while(isspace(*s))
		s++;
	t = strchr(s, 0);
	t--;
	while(isspace(*t ))
		t--;
	t++;
	*t = 0;
	return s;
}

/*	"/proc" is required for postnote	*/

static int
notefun(void *a, char *msg)
{
	char buf[32];
	int fd;
	USED(a);
	/* 	postnote(PNGROUP, cpid, msg);
	*	we use #p		*/
	snprint(buf,sizeof buf,"#p/%d/notepg",cpid);
	fd = open(buf, OWRITE);
	if(fd<0)
		syslog(0,HTTPLOG, "notefun: open %s : %r");
	else{
		write(fd,msg,strlen(msg));
		close(fd);
	}
	if(strcmp(msg, "alarm") == 0)
		return 1;
	return 0;
}


/*	merge dirs bellow mnt to root
*	the bind option is:
*	doc and etc are MREPL others are MAFTER
*	-1 is returned if doc is absent
*	
*	currently only one directory level is supported
*	this limitation should be removed in future.
*	example
*	let dirs in mnt be:
*		mnt/doc
*		mnt/etc
*		mnt/data
*	and dirs in root be:
*		root/doc
*		root/lib
*	then is routine designed to do:
*		bind mnt/doc root/doc
*		bind -a mnt root
*	-Kenar-	*/
static int
merge(char *mnt, char *root)
{
	int fd, n, i, status, f;
	Dir dir[10], dir1,dir2;
	char *name, path1[256], path2[256];
	f = 0;
	fd = open(mnt,OREAD);
	if(fd < 0){
		syslog(0,HTTPLOG,"error: not open %s : %r", mnt);
		return -1;
	}
	n = dirread(fd, dir, sizeof dir)/sizeof(Dir);
	for(i = 0; i < n; i++){
		name = dir[i].name;
		snprint(path1, sizeof path1, "%s/%s",mnt, name);
		status = dirstat(path1,&dir1);
		if(!status && (dir1.mode & CHDIR)){
			snprint(path2, sizeof path2, "%s/%s", root, name);
			status = dirstat(path2, &dir2);
			if(!status && (dir2.mode & CHDIR)){
				if(strcmp(name,"doc") == 0 || strcmp(name,"etc") == 0){
					status = bind(path1, path2, MREPL);
					if(strcmp(name,"doc") == 0) f++;
				}
				else
					status = bind(path1, path2, MAFTER);
				if(status < 0){
					syslog(0,HTTPLOG,"error: bind %s %s : %r", path1, path2);
					return -1;
				}
			}
		}
	}
	if(f == 0){
		syslog(0,HTTPLOG,"error: no doc in %s", mnt);
		return -1;
	}
	status = bind(webmnt, webroot, MAFTER);
	if(status < 0){
		syslog(0,HTTPLOG,"error: bind %s %s : %r", webmnt, webroot);
		return -1;
	}
	return 0;
}

/*
*	shell style matching function
*	I modified to simplify original one
*	ref: /sys/src/cmd/rc/glob.c
*	-Kenar-
*
 * 	Does the string s match the pattern p
 * 	* matches any sequence of characters
 * 	? matches any single character
 * 	[...] matches the enclosed list of characters
 *	~ in the beginning of [...] means compliment
 *	- in [...] means range in ascii order
 *
 *	return value
 *	1: matched
 *	0: not matched
 */
static int
match(char *s, char *p)
{
	int compl, hit, lo, hi, t, c;
	for(;*p; s++,p++){
		switch(*p){
		case '*':
			++p;
			for(;;){
				if(match(s, p)) return 1;
				if(!*s) break;
				s++;
			}
			return 0;
		case '?':
			if(*s=='\0') return 0;
			break;
		case '[':
			if(*s=='\0') return 0;
			c=*s;
			p++;
			compl=*p=='~';
			if(compl) p++;
			hit=0;
			while(*p!=']'){
				if(*p=='\0') return 0;		/* syntax error */
				lo=*p;
				p++;
				if(*p!='-') hi=lo;
				else{
					p++;
					if(*p=='\0') return 0;	/* syntax error */
					hi=*p;
					p++;
					if(hi<lo){ t=lo; lo=hi; hi=t; }
				}
				if(lo<=c && c<=hi) hit=1;
			}
			if(compl) hit=!hit;
			if(!hit) return 0;
			break;
		default:
			if(*p != *s) return 0;
		}
	}
	return *s=='\0';
}


static char *
skip(char *s, char *p)
{
	while(*s && strchr(p, *s))
		s++;
	return s;
}

/*	getword get a word from s and returns point
*	next to search ( the point is next char of the word)
*	if there is no word, it returns nil
*	-Kenar-	*/
static char *
getword(char *s, char *w, int size)
{
	char *r;
	r = w + size;
	s = skip(s, " \t");
	if(*s == 0)
		return nil;
	while(*s && !strchr(" \t", *s)){
		if(w == r){
			w--;
			break;
		}
		*w++ = *s++;
	}
	*w = 0;
	return s;
}

/*
*	The format of "/etc/allow" is shown using example:
*	/lecture/for/our/students
*		192.168.[1-8].*
*		...
*	path
*		...
*	where ip is given by shell style pattern matching.
*
*	restrictions := restriction restriction ...
*	restriction := pathline iplines
*	pathline := path
*	iplines := ipline ipline ..
*	ipline := spaces ippattern ippattern ..
*
*	# return values
*	0: the IP is allowed
*	-1: the IP is disallowed
*/
static int
ipcheck(HConnect *c, char *file, char *ip)
{
	char *path, *p, *line, *s;
	int ok, check;
	Biobuf *b;
	char word[64];

	//logit(c,"ipcheck: ip=%s", ip);

	path = c->req.path;
	b = Bopen(file, OREAD);
	if(b){
		ok = 1;
		check = 0;
		while((line = Brdline(b, '\n')) != nil){
			line[Blinelen(b)-1] = 0;
			s = strchr(line, '#');
			if(s != nil)
				*s = '\0'; 	/* chop comment */
			/* skip null line */
			if(*skip(line," \t") == 0)
				continue;

			if(strchr(" \t", line[0])){	// ip pattern of lpath
				if(check == 0) // skip the ip pattern
					continue;
				if(ok)
					continue;
				/* 	how can I get single pattern ?
				*	ip pattern has no space	OK */
				while(line = getword(line, word, sizeof word)){ // asign =
					//logit(c,"ipcheck: word=%s", word);
					if(word[0] == '~' && match(ip, word+1) == 0){
						ok = 1;
						break;
					}
					if(match(ip, word)){
						ok = 1;
						break;
					}
				}
				if(!ok) // examine more ip line
					continue;
			} else {	// path
				/* first we must check previous analysys */
				if(check && !ok)
					break;
				/* start new path line */
				p = strpbrk(line," \t");
				if(p)
					*p = 0;
				/* 	check some human error, if error put some
				*	guards and let that know. */
				if(line[0] != '/'){
					logit(c, "warning: path in /etc/allow must begin with '/': %s",line);
					path++;
				}
				if(strncmp(line, path, strlen(line)) == 0){
					/* 	then we must confirm
					*	the client IP is allowed to see our data */
					// logit(c,"line=%s path=%s", line, path);
					check = 1;
					ok = 0;
				}
				else{
					check = 0;
					ok = 1;
				}
			}
		}
		Bterm(b);
		return  ok - 1;
	}
	return 0;
}

#ifdef NODEF
static int
ipcheck(HConnect *c, char *file, char *ip)
{
	char *path, *p, *line, *s;
	int ok;
	Biobuf *b;
	char lpath[256], word[64];

	//logit(c,"ipcheck: ip=%s", ip);

	path = c->req.path;
	b = Bopen(file, OREAD);
	if(b){
		ok = 1;
		lpath[0] = 0;
		while((line = Brdline(b, '\n')) != nil){
			line[Blinelen(b)-1] = 0;
			s = strchr(line, '#');
			if(s != nil)
				*s = '\0'; 	/* chop comment */
			/* skip null line */
			if(*skip(line," \t") == 0)
				continue;

			if(strchr(" \t", line[0])){	// ip pattern of lpath
				if(lpath[0] == 0) // skip the ip pattern
					continue;
				if(ok)
					continue;
				/* 	how can I get single pattern ?
				*	ip pattern has no space	OK */
				while(line = getword(line, word, sizeof word)){ // asign =
					//logit(c,"ipcheck: word=%s", word);
					if(word[0] == '~' && match(ip, word+1) == 0){
						ok = 1;
						break;
					}
					if(match(ip, word)){
						ok = 1;
						break;
					}
				}
				if(!ok) // examine more ip line
					continue;
			} else {	// path
				/* first we must check previous analysys */
				if(lpath[0] && !ok)
					break;
				/* start new path line */
				p = strpbrk(line," \t");
				if(p)
					*p = 0;
				/* 	check some human error, if error put some
				*	guards and let that know. */
				if(line[0] != '/'){
					logit(c, "warning: path in /etc/allow must begin with '/': %s",line);
					path++;
				}
				if(strncmp(line, path, strlen(line)) == 0){
					/* 	then we must confirm
					*	the client IP is allowed to see our data */
					// logit(c,"line=%s path=%s", line, path);
					strncpy(lpath, line, sizeof lpath);
					ok = 0;
				}
				else{
					lpath[0] = 0;
					ok = 1;
				}
			}
		}
		Bterm(b);
		return  ok - 1;
	}
	return 0;
}
#endif

/*	path redirection for users and virtual hosts
*	this redirection is used after namespace configuration
*
*	# return values
*	nil: not redirected
*	else: new uri
*	-Kenar-	*/
static char*
redirect(HConnect *c, char *file, char *path)
{
	char *pat, *rep, *line, *s, *buf, *user;
	Biobuf *b;
	char *field[2];
	int m, m0, bufsize;

	b = Bopen(file, OREAD);
	if(b){
		pat = nil;
		rep = nil;
		m0 = 0;
		while((line = Brdline(b, '\n')) != nil){
			line[Blinelen(b)-1] = 0;
			s = strchr(line, '#');
			if(s != nil)
				*s = '\0'; 	/* chop comment */
			/* skip null line */
			if(*skip(line," \t") == 0)
				continue;

			if(getfields(line, field, 2, 1, " \t") != 2)
				continue;
			m = strlen(field[0]);
			if(strncmp(field[0], path, m) == 0){
				if(m >= m0){
					m0 = m;
					pat = field[0];
					rep = field[1];
				}
			}
		}
		if(m0 > 0){
			if(rep[0] == '/' ) {
				/* must be a path name of this httpd name space */
				user = c->req.user;
				if(user){
					bufsize = strlen(user) + strlen(rep)
						+ strlen(path) - strlen(pat) + strlen("/~") + 1 + UTFmax;
					buf = halloc(c, bufsize);
					snprint(buf, bufsize, "/~%s%s%s", user, rep, path + strlen(pat));
				} else {
					bufsize = strlen(rep)
						+ strlen(path) - strlen(pat) + 1 + UTFmax;
					buf = halloc(c, bufsize);
					snprint(buf, bufsize, "%s%s", rep, path + strlen(pat));
				}
			}
			else{ // assumed "http://...."
				bufsize = strlen(rep) + strlen(path) - strlen(pat) + 1 + UTFmax;
				buf = halloc(c, bufsize);
				snprint(buf, bufsize, "%s%s", rep, path + strlen(pat));
			}
			rep = buf;
		}
		Bterm(b);
		return rep;
	}
	return nil;
}
